/******************************************************************************
* Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
* - Redistributions of source code must retain the above copyright notice, this
*   list of conditions and the following disclaimer.
* - Redistributions in binary form must reproduce the above copyright notice,
*   this list of conditions and the following disclaimer in the documentation
*   and/or other materials provided with the distribution.
* - Modified versions of the software must be conspicuously marked as such.
* - This software is licensed solely and exclusively for use with
*   processors/products manufactured by or for Analog Devices, Inc.
* - This software may not be combined or merged with other code in any manner
*   that would cause the software to become subject to terms and conditions
*   which differ from those listed here.
* - Neither the name of Analog Devices, Inc. nor the names of its contributors
*   may be used to endorse or promote products derived from this software
*   without specific prior written permission.
* - The use of this software may or may not infringe the patent rights of one
*   or more patent holders.  This license does not release you from the
*   requirement that you obtain separate licenses from these patent holders to
*   use this software.
*
* THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
* NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
* CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
******************************************************************************/
#include <iostream>
#include <string>
#include "stdint.h"

#ifdef SWIG_PY
#include "Python.h"
#endif // SWIG_PY

#include "watch.hpp"
#include "git_version.hpp"
#include "byte_swap.hpp"

#define MAXIMUM_FILE 64
#define BATTERY_LEVEL 5
#define MEMORY_REACHED 5
#define Debug 0

/*!
   \brief Constructor for the top-level Watch class
*/
watch::watch(watch_phy_callback &cb):
      pkt_phy_writer(cb),
      adxl_app(this),
			eda_app(this),
      ecg_app(this),
    ad5940_app(this),
    temperature_app(this),
      adpd_app(this),
			ppg_app(this),
			syncppg_app(this),
			adpd4000_app(this),
      pm_app(this),
      fs_app(this){}


const std::string watch::SDK_version = "3.0.3";
const std::string watch::git_hash = g_GIT_HASH;
const std::string watch::git_describe = g_GIT_DESCRIBE;

/*!
   \brief Perform a basic self-test
*/
void watch::self_test(void) {
  std::cout << "=======================================================" << std::endl;
  std::cout << "SELF TEST" << std::endl;
  std::cout << "=======================================================" << std::endl;
  std::cout << "Welcome to ADI's Watch SDK" << std::endl;
  std::cout << "Version: " << this->SDK_version << std::endl;
  //std::cout << "Git hash: " << this->git_hash << std::endl;
  //std::cout << "Git describe: " << this->git_describe << std::endl;
}

/*!
   \brief set the platform. @param platForm (python/windows/android/ios)
   Note:Platform must be set before executing any command
*/
void watch::set_platform(platforms platForm) {
  platform = platForm;
}

/*!
   \brief Returns a string describing the version of the SDK.
*/
std::string watch::get_version(void) {
  return this->git_hash;
}

/*!
\brief Call physical callback sys_alert_method with the system error message
*/
void watch::system_alert_call(uint8_t alert_message)
{
	this->pkt_phy_writer.sys_alert_call(ADI_SDK_PACKET_TIMED_OUT);
}

/*!
   \brief Called by Watch application objects to send a packet over a physical interface
*/
void watch::write_bytes(std::vector<uint8_t> &bytes) {
#ifdef SWIG_PY
// We need to make sure that this thread is registered with Python's Global Interpreter Lock before calling out to Python code
  PyGILState_STATE gstate;
  gstate = PyGILState_Ensure();
#endif // SWIG_PY

  m2m2_hdr_t * p_hdr = (m2m2_hdr_t*)&bytes[0];
  m2m2_pkt<_m2m2_app_common_cmd_t> pkt;
  pkt.unpack(bytes);
  p_hdr->src = this->get_address();
  // Swap the endianness of the header fields
  // src
  std::swap(bytes[0], bytes[1]);
  // dest
  std::swap(bytes[2], bytes[3]);
  // length
  std::swap(bytes[4], bytes[5]);
  // TODO: calculate checksum
  // Call the PHY callback
  this->pkt_phy_writer.call(bytes);


#ifdef SWIG_PY
// Clean up Python's GIL lock
  PyGILState_Release(gstate);
#endif // SWIG_PY
}

/*!
   \brief Returns the M2M2 address assigned to the SDK.
*/
M2M2_ADDR_ENUM_t watch::get_address(void) {
  switch (platform) {
    case python:
      return M2M2_ADDR_APP_CLI;
    case windows:
      return M2M2_ADDR_APP_WT;
    case android:
      return M2M2_ADDR_APP_DROID;
    case ios:
      return M2M2_ADDR_APP_IOS;
    default:
      return M2M2_ADDR_UNDEFINED;
  }
}


/*!
   \brief Dispatches a packet of bytes to the appropriate application objects within the SDK.

   Call this function whenever you receive bytes over the physical interface from the device.
*/
void watch::dispatch(std::vector<uint8_t> bytes) {
  m2m2_hdr_t *hdr = reinterpret_cast<m2m2_hdr_t *> (&bytes[0]);
#if Debug
  std::cout << "C++ return: Received a Packet " << BYTE_SWAP_16(hdr->src) <<std::endl;
#endif
  switch(BYTE_SWAP_16(hdr->src)) {
    case M2M2_ADDR_SENSOR_ADXL:
      this->adxl_app.dispatch(bytes);
      break;
    case M2M2_ADDR_SENSOR_ADXL_STREAM:
      this->adxl_app.adxl_stream.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_ECG:
      this->ecg_app.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_EDA:
      this->eda_app.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_EDA_STREAM:
      this->eda_app.eda_stream.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_ECG_STREAM:
      this->ecg_app.ecg_stream.dispatch(bytes);
      break;
    case M2M2_ADDR_SENSOR_AD5940:
      this->ad5940_app.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_TEMPERATURE:
      this->temperature_app.dispatch(bytes);
      break;
    case M2M2_ADDR_MED_TEMPERATURE_STREAM:
      this->temperature_app.temperature_stream.dispatch(bytes);
      break;
    case M2M2_ADDR_SENSOR_ADPD_STREAM:
#if Debug
		std::cout << "C++ return: Received a Packet : M2M2_ADDR_SENSOR_ADPD_STREAM!" << std::endl;
#endif
      this->adpd_app.adpd_stream.dispatch(bytes);
      break;
    case M2M2_ADDR_SENSOR_ADPD:
      this->adpd_app.dispatch(bytes);
      break;
		case M2M2_ADDR_MED_PPG:
			this->ppg_app.dispatch(bytes);
			break;
		case M2M2_ADDR_MED_PPG_STREAM:
			this->ppg_app.ppg_stream.dispatch(bytes);
			break;
		case M2M2_ADDR_MED_SYNC_ADPD_ADXL:
			this->syncppg_app.dispatch(bytes);
			break;
		case M2M2_ADDR_MED_SYNC_ADPD_ADXL_STREAM:
			this->syncppg_app.syncppg_stream.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPDCL:
			this->adpd4000_app.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM1:
			this->adpd4000_app.adpd4000_stream1.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM2:
			this->adpd4000_app.adpd4000_stream2.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM3:
			this->adpd4000_app.adpd4000_stream3.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM4:
			this->adpd4000_app.adpd4000_stream4.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM5:
			this->adpd4000_app.adpd4000_stream5.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM6:
			this->adpd4000_app.adpd4000_stream6.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM7:
			this->adpd4000_app.adpd4000_stream7.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM8:
			this->adpd4000_app.adpd4000_stream8.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM9:
			this->adpd4000_app.adpd4000_stream9.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM10:
			this->adpd4000_app.adpd4000_stream10.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM11:
			this->adpd4000_app.adpd4000_stream11.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_STREAM12:
			this->adpd4000_app.adpd4000_stream12.dispatch(bytes);
			break;
		case M2M2_ADDR_SENSOR_ADPD_OPTIONAL_BYTES_STREAM:
			this->adpd4000_app.adpd4000_optional_stream.dispatch(bytes);
			break;
		case M2M2_ADDR_SYS_FS:
			if (bytes[8] == M2M2_FILE_SYS_CMD_LS_RESP) {
				this->fs_app.fs_ls_response.dispatch(bytes);
			}
			else  if (bytes[8] == M2M2_FILE_SYS_CMD_GET_RESP) {
				this->fs_app.fs_stream.dispatch(bytes);
			}
			else {
				this->fs_app.dispatch(bytes);
			}
			break;
		case M2M2_ADDR_SYS_PM:
			this->pm_app.dispatch(bytes);
			break;
		case M2M2_ADDR_SYS_BATT_STREAM:
			this->pm_app.battery_stream.dispatch(bytes);
			break;
		default:
			std::cout << "C++::WARNING: Received a Packet that couldn't be dispatched!" << std::endl;
			std::cout << "    Packet is:" << std::endl;
			for (unsigned int i = 0; i < bytes.size(); i++) {
				std::cout << " " << std::hex << std::setfill('0') << std::setw(2) << (int) bytes[i];
			}
			std::cout << std::endl;
  }
}


/**
* Start the PPG, sync-adpd-adxl and AGC stream with a default configuration.
* @param ppg_stream_cb ppg stream callback to be registered
* @param sync_stream_cb sync ppg stream callback to be registered
* @param agc_stream_cb agc stream callback to registered
* @param enableAGC set true to enable AGC stream
* @return returns the status
*/
ret::err_status watch::start_syncppg(ppg_stream_callback &ppg_stream_cb, syncppg_stream_callback &sync_stream_cb) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;
	status = this->adpd_app.load_cfg();
	if (status == ret::SDK_OK)
	{
		status = this->ppg_app.load_lcfg();
	}
	else
	{
		return ret::LOAD_CFG_FAILED;

	}

	if (status == ret::SDK_OK)
	{
		status = this->adpd_app.calibrate_clock();
	}
	else
	{
		return ret::LOAD_CFG_FAILED;
	}

	if (status == ret::SDK_OK)
	{
		status = this->ppg_app.ppg_stream.subscribe(ppg_stream_cb);
	}
	else
	{
		return ret::CLOCK_CALIBRATE_FAILED;
	}

	if (status == ret::SDK_OK)
	{
		status = this->syncppg_app.syncppg_stream.subscribe(sync_stream_cb);
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}

	if (status == ret::SDK_OK)
	{
		stream_status = this->ppg_app.ppg_stream.start();
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
		return ret::STREAM_START_FAILED;
	}
	else {
		return ret::STREAM_STARTED;
	}
	
}

/**
* Stop the PPG, sync-adpd-adxl and AGC streams
* @param disableAGC set true to stop AGC stream
* @return returns the status
*/
ret::err_status watch::stop_syncppg(void) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	stream_status = this->ppg_app.ppg_stream.stop();
	status = this->ppg_app.ppg_stream.unsubscribe();
	status = this->syncppg_app.syncppg_stream.unsubscribe();



	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
	}
	else {
		status = this->ppg_app.ppg_stream.unsubscribe();
	}

	if (status == ret::SDK_OK)
	{
		status = this->syncppg_app.syncppg_stream.unsubscribe();

	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}

	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STOPPED;
	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}

}


/*!
   \brief Start the ADPD stream with default configuration.
*/
ret::err_status watch::start_adpd(adpd_stream_callback &stream_cb) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

  std::chrono::steady_clock::time_point begin = std::chrono::steady_clock::now();
  status = this->adpd_app.load_cfg();
  if (status == ret::SDK_OK)
  {
	  status = this->adpd_app.calibrate_clock();
  }
  else
  {
	  return ret::LOAD_CFG_FAILED;

  }
  if (status == ret::SDK_OK)
  {
	  status = this->adpd_app.set_slot(this->adpd_app.slotA, this->adpd_app.slotB);
  }
  else
  {
	  return ret::CLOCK_CALIBRATE_FAILED;
  }
  if (status == ret::SDK_OK)
  {
	  status = this->adpd_app.adpd_stream.subscribe(stream_cb);

  }
  else
  {
	  return ret::SET_SLOT_FAILED;
  }
  if (status == ret::SDK_OK)
  {
	  stream_status = this->adpd_app.adpd_stream.start();
  }
  else
  {
	  return ret::SUBSCRIBE_FAILED;
  }
  if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
	  return ret::STREAM_START_FAILED;
  }
  else {
	  return ret::STREAM_STARTED;
  }

  std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
#if Debug
  std::cout << "SDK:: Time difference in start_adpd= " << std::chrono::duration_cast<std::chrono::milliseconds>(end - begin).count() << std::endl;
#endif
}

/*!
   \brief Stop the ADPD stream.
*/
ret::err_status watch::stop_adpd(void) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	std::chrono::steady_clock::time_point begin = std::chrono::steady_clock::now();

	stream_status = this->adpd_app.adpd_stream.stop();

  if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
  }
  else {
		status = this->adpd_app.adpd_stream.unsubscribe();
  }

  if (status == ret::SDK_OK)
  {
	  return ret::STREAM_STOPPED;
  }
  else
  {
	  return ret::UNSUBSCRIBE_FAILED;
  }

  std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
#if Debug
  std::cout << "SDK:: Time difference in stop_adpd= " << std::chrono::duration_cast<std::chrono::milliseconds>(end - begin).count() << std::endl;
#endif
}
/*!
   \brief Start the temperature stream with a default configuration.
*/
ret::err_status watch::start_temperature(temperature_stream_callback &stream_cb) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_start_status;
	stream_start_status = this->temperature_app.temperature_stream.start();
	if (stream_start_status == m2m2_application::APP_COMMON_STATUS_ERROR ||
		stream_start_status == m2m2_application::APP_COMMON_STATUS_STREAM_NOT_STARTED) {
		return ret::STREAM_START_FAILED;
	}
	else
	{
		//For sensor board 3, AD7689 is not required to start temperature
		status = this->temperature_app.temperature_stream.subscribe(stream_cb);
	}

	if (status == ret::SDK_OK) {
		return ret::STREAM_STARTED;
	}
	else {
		return ret::SUBSCRIBE_FAILED;
	}

}

/*!
   \brief Stop the temperature stream.
*/
ret::err_status watch::stop_temperature(void) {
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_stop_status;
	ret::sdk_status status;
	stream_stop_status = this->temperature_app.temperature_stream.stop();
	if (stream_stop_status == m2m2_application::APP_COMMON_STATUS_ERROR ||
		stream_stop_status == m2m2_application::APP_COMMON_STATUS_STREAM_NOT_STOPPED) {
		return ret::STREAM_STOP_FAILED;
	}
	else
	{
		//For sensor board 3, AD7689 is not required to stop temperature
		status = this->temperature_app.temperature_stream.unsubscribe();
	}

	if (status == ret::SDK_OK) {
		return ret::STREAM_STOPPED;
	}
	else {
		return ret::UNSUBSCRIBE_FAILED;
	}

}
/*!
   \brief Start the ADXL stream with a default configuration.
*/
ret::err_status watch::start_adxl(adxl_stream_callback &stream_cb) {
  ret::sdk_status status;
  m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

  status = this->adxl_app.adxl_stream.subscribe(stream_cb);
	if (status == ret::SDK_OK)
	{
		stream_status = this->adxl_app.adxl_stream.start();
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
		return ret::STREAM_START_FAILED;
	}
	else {
		return ret::STREAM_STARTED;
	}

}

/*!
   \brief Stop the ADXL stream.
*/
ret::err_status watch::stop_adxl(void) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	stream_status = this->adxl_app.adxl_stream.stop();

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
	}
	else {
		status = this->adxl_app.adxl_stream.unsubscribe();
	}
	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STOPPED;
	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}
  
}

/*!
\brief Start the ADPD4000 stream with default configuration.
*/
ret::err_status watch::start_adpd4000(adpd4000_stream_callback &stream_cb) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	this->adpd4000_app.adpd4000_stream1.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream2.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream3.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream4.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream5.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream6.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream7.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream8.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream9.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream10.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream11.subscribe(stream_cb);
	this->adpd4000_app.adpd4000_stream12.subscribe(stream_cb);
	stream_status =  this->adpd4000_app.adpd4000_stream1.start();

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
		return ret::STREAM_START_FAILED;
}
	else {
		return ret::STREAM_STARTED;
	}


	std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();
#if Debug
	std::cout << "SDK:: Time difference in start_adpd= " << std::chrono::duration_cast<std::chrono::milliseconds>(end - begin).count() << std::endl;
#endif
}

/*!
\brief Stop the ADPD4000 stream.
*/
ret::err_status watch::stop_adpd4000(void) {
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	std::chrono::steady_clock::time_point begin = std::chrono::steady_clock::now();
	stream_status = this->adpd4000_app.adpd4000_stream1.stop();
	this->adpd4000_app.adpd4000_stream1.unsubscribe();
	this->adpd4000_app.adpd4000_stream2.unsubscribe();
	this->adpd4000_app.adpd4000_stream3.unsubscribe();
	this->adpd4000_app.adpd4000_stream4.unsubscribe();
	this->adpd4000_app.adpd4000_stream5.unsubscribe();
	this->adpd4000_app.adpd4000_stream6.unsubscribe();
	this->adpd4000_app.adpd4000_stream7.unsubscribe();
	this->adpd4000_app.adpd4000_stream8.unsubscribe();
	this->adpd4000_app.adpd4000_stream9.unsubscribe();
	this->adpd4000_app.adpd4000_stream10.unsubscribe();
	this->adpd4000_app.adpd4000_stream11.unsubscribe();
	this->adpd4000_app.adpd4000_stream12.unsubscribe();
	
	std::chrono::steady_clock::time_point end = std::chrono::steady_clock::now();

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
	}
	else {
		return ret::STREAM_STOPPED;
	}


#if Debug
	std::cout << "SDK:: Time difference in stop_adpd= " << std::chrono::duration_cast<std::chrono::milliseconds>(end - begin).count() << std::endl;
#endif
}


ret::err_status watch::start_battery(battery_stream_callback &stream_cb) {
	ret::sdk_status status;
  status = this->pm_app.battery_stream.subscribe(stream_cb);
	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STARTED;
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}
}

ret::err_status watch::stop_battery(void) {
	ret::sdk_status status;
  status = this->pm_app.battery_stream.unsubscribe();
	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STOPPED;
	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}
}


ret::err_status watch::start_ecg(ecg_stream_callback &stream_cb)
{
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	std::vector<std::pair<uint8_t, uint16_t>> addr_vals;
	std::vector<std::pair<uint8_t, uint16_t>> ret_vals;
	addr_vals.push_back(std::make_pair(static_cast<uint8_t>(0), static_cast<uint16_t>(100)));
	addr_vals.push_back(std::make_pair(static_cast<uint8_t>(2), static_cast<uint16_t>(1)));
	addr_vals.push_back(std::make_pair(static_cast<uint8_t>(5), static_cast<uint16_t>(0)));
	ret_vals = this->ecg_app.lcfg_write(addr_vals);

	if (ret_vals.size() == 3 && ret_vals[0].second == 100 && ret_vals[1].second == 1 && ret_vals[2].second == 0) {
	  status = this->ad5940_app.start();
	}
	else
	{
		return ret::LCFG_WRITE_FAILED;
	}
	if (status == ret::SDK_OK)
	{
		stream_status = this->ecg_app.ecg_stream.start();
	}
	else
	{
		return ret::STREAM_START_FAILED;
	}

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
		return ret::STREAM_START_FAILED;
	}
	else {
		status = this->ecg_app.ecg_stream.subscribe(stream_cb);
	}
	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STARTED;
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}

}
ret::err_status watch::stop_ecg(void)
{
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	status =	this->ad5940_app.stop();

	if (status == ret::SDK_OK)
	{
		stream_status = this->ecg_app.ecg_stream.stop();
	}
	else
	{
		return ret::STREAM_START_FAILED;
	}


	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
	}
	else {
		status = this->ecg_app.ecg_stream.unsubscribe();
	}

	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STOPPED;
	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}
	
}

ret::err_status watch::start_eda(eda_stream_callback &stream_cb)
{
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	status = this->ad5940_app.start();

	if (status == ret::SDK_OK)
	{
		stream_status = this->eda_app.eda_stream.start();
	}
	else
	{
		return ret::STREAM_START_FAILED;
	}

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STARTED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_IN_PROGRESS) {
		return ret::STREAM_START_FAILED;
	}
	else {
		status = this->eda_app.eda_stream.subscribe(stream_cb);
	}
	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STARTED;
	}
	else
	{
		return ret::SUBSCRIBE_FAILED;
	}	
	
}
ret::err_status watch::stop_eda(void)
{
	ret::sdk_status status;
	m2m2_application::APP_COMMON_STATUS_ENUM_t stream_status;

	status = this->ad5940_app.stop();

	if (status == ret::SDK_OK)
	{
		stream_status = this->eda_app.eda_stream.stop();
	}
	else
	{
		return ret::STREAM_START_FAILED;
	}

	if (stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_STOPPED && stream_status != m2m2_application::APP_COMMON_STATUS_STREAM_COUNT_DECREMENT) {
		return ret::STREAM_STOP_FAILED;
	}
	else {
		status = this->eda_app.eda_stream.unsubscribe();
	}

	if (status == ret::SDK_OK)
	{
		return ret::STREAM_STOPPED;
	}
	else
	{
		return ret::UNSUBSCRIBE_FAILED;
	}

}


/*!
 \brief Start the Logging of ADXL stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::start_log_adxl(void) {
  this->fs_app.fs_subscribe(fs_application::ADXL_Stream);
  m2m2_application::APP_COMMON_STATUS_ENUM_t status = this->fs_app.fs_stream.start();
  this->adxl_app.adxl_stream.start();
  return status;

}

/*!
 \brief Stop the Logging of ADXL stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::stop_log_adxl(void) {
  this->adxl_app.adxl_stream.stop();
  this->fs_app.fs_unsubscribe(fs_application::ADXL_Stream);
  return this->fs_app.fs_stream.stop();
}


/*!
 \brief Start the logging of ADPD stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::start_log_adpd(void) {
  this->adpd_app.load_cfg();
  this->adpd_app.calibrate_clock();
  this->adpd_app.set_slot(adpd_application::SLOTMODE_DISABLED,
                          adpd_application::SLOTMODE_SUM32b);
  this->fs_app.fs_subscribe(fs_application::ADPD_Stream);
  m2m2_application::APP_COMMON_STATUS_ENUM_t status = this->fs_app.fs_stream.start();
  this->adpd_app.adpd_stream.start();
  return status;

}

/*!
 \brief Stop the logging of ADPD stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::stop_log_adpd(void) {
  this->adpd_app.adpd_stream.stop();
  this->fs_app.fs_unsubscribe(fs_application::ADPD_Stream);
  return this->fs_app.fs_stream.stop();

}

/*!
 \brief Start the logging of battery stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::start_log_battery(void) {
  this->fs_app.fs_subscribe(fs_application::BATTERY_Stream);
  m2m2_application::APP_COMMON_STATUS_ENUM_t status = this->fs_app.fs_stream.start();
  return status;

}

/*!
 \brief Stop the logging of battery stream.
 */
m2m2_application::APP_COMMON_STATUS_ENUM_t watch::stop_log_battery(void) {

 this->fs_app.fs_unsubscribe(fs_application::BATTERY_Stream);
 return this->fs_app.fs_stream.stop();
  
}
